/*
 * Copyright 2007 Sun Microsystems, Inc.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Sun Microsystems nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package demo.alarm;

import com.sun.scenario.animation.Clip;
import com.sun.scenario.animation.TimingTargetAdapter;
import com.sun.scenario.scenegraph.JSGPanel;
import com.sun.scenario.scenegraph.SGGroup;
import com.sun.scenario.scenegraph.SGNode;
import com.sun.scenario.scenegraph.SGShape;
import com.sun.scenario.scenegraph.SGText;
import com.sun.scenario.scenegraph.SGText.HAlign;
import com.sun.scenario.scenegraph.SGText.VAlign;
import com.sun.scenario.scenegraph.SGTransform;
import com.sun.scenario.scenegraph.event.SGMouseAdapter;
import com.sun.scenario.scenegraph.fx.FXShape;
import com.sun.scenario.scenegraph.fx.FXText;
import java.applet.Applet;
import java.applet.AudioClip;
import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.MultipleGradientPaint.CycleMethod;
import java.awt.Paint;
import java.awt.RadialGradientPaint;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.net.URL;
import java.util.concurrent.TimeUnit;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

/**
 * An egg timer example where you can set for up to 59 minutes. When it is ringing 
 * at zero, it will be automatically off after 1 minute unless the red led 
 * is pressed. 
 * 
 * Usage:
 * <ul>
 * <li>Turn On: Set the knob to the minutes you want. The red led will blink.</li>
 * <li>Turn Off: Press the red led indicator on the knob. If the alarm is ringing, 
 * it will be off.</li>
 * <li>Resume: Press the knob.</li>
 * </ul>
 * 
 * @author Baechul Kim
 */
public class EggTimer extends JSGPanel {

    private static JFrame frame;
    private Clip time;
    /**
     * The radius of outmost circle in the egg timer. There are 3 circles, one for
     * labels, others for ticks and a disk(inner most one) where the knob is 
     * rotating. 
     */
    private double r;
    /**
     * The radius difference of the circles.
     */
    private double m;
    /**
     * A led to indicate if the egg timer is on. It is also used as a button to 
     * pause the alarm sound. Pressing the knob handle will 
     * resume the timer.
     */
    private FXShape led;
    private boolean reset;
    private int cx,  cy;
    private int alarmStart;
    /**
     * The "danger.au" originally from a demo in the JDK bundle. 
     */
    private AudioClip defaultSound;
    // UI decoration
    private int fontSize;
    private SGTransform.Affine knobAffine;
    private SGTransform.Affine refAffine;
    private Color[] ledColors = {new Color(0x3F0808), Color.red, Color.yellow};
    private Paint knobFillColor = new Color(0x8AA975);
    private Paint diskFillColor = new Color(0xC7C9C6);

    public EggTimer() {
        setBackground(new Color(1.0f, 1.0f, 1.0f, 0.5f));

        ////////////////////////////////////////////////////////////////////////
        // Initialization
        setToolTipText("Turn or press the knob to start. Press the red indicator to stop. A popup menu for others");
        r = 100;
        m = r / 5;
        reset = true;
        alarmStart = 1;

        if (r < 60) {
            fontSize = 10;
        } else if (r > 199) {
            fontSize = 24;
        } else {
            fontSize = 10 + (int) ((r - 60) * 14 / 140);
        }

        Point2D center = new Point2D.Double(r, r);
        float radius = (float) r / 2;
        Point2D focus = new Point2D.Double(r / 2, r / 2);
        float[] dist = {0.1f, 1.0f};
        Color[] colors = {Color.white, (Color) knobFillColor};
        knobFillColor = new RadialGradientPaint(center, radius, focus,
                dist, colors, CycleMethod.NO_CYCLE);

        radius = (float) r;
        focus = new Point2D.Double(1.5 * r, 1.5 * r);
        float[] dist2 = {0.1f, 1.0f};
        Color[] colors2 = {Color.white, (Color) diskFillColor};
        diskFillColor = new RadialGradientPaint(center, radius, focus,
                dist2, colors2, CycleMethod.NO_CYCLE);

        URL url = this.getClass().getResource("resources/danger.au");
        defaultSound = Applet.newAudioClip(url);

        ////////////////////////////////////////////////////////////////////////        
        invalidateScene();

        time = Clip.create(Clip.INDEFINITE, Clip.INDEFINITE,
                new TimingTargetAdapter() {
            
                    private long soundOnTime = Long.MIN_VALUE;
                    private long start = Long.MIN_VALUE;
                    private long lastUpdate;
                    private boolean alarmStartFlag = false;
                    private boolean alarmEndFlag = false;

                    @Override
                    public void timingEvent(float t, long totalElapsed) {
                        if (reset) {
                            reset = false;
                            soundOnTime = Long.MIN_VALUE;
                            start = Long.MIN_VALUE;
                            alarmStartFlag = false;
                            alarmEndFlag = false;
                        }

                        if (start == Long.MIN_VALUE) {
                            start = System.nanoTime();
                            lastUpdate = start;
                        }

                        // update twice per second.
                        long now = System.nanoTime();
                        if (TimeUnit.NANOSECONDS.toMillis(now - lastUpdate) > 500) {
                            // moves 6 degree per minute
                            double dt = TimeUnit.NANOSECONDS.toMillis(now - start);
                            dt = (60 / (double) 60000) * dt;
                            rotateBy(-1 * Math.toRadians(dt));
                            start = System.nanoTime();
                            lastUpdate = start;

                            // Dead zone: -0.5 to +0.5. Treated like 0 degree.
                            double a = getKnobAngle();
                            if (a > 359.0 || a < 0.95) {
                                a = 0;
                                rotate(0);
                            }
                            if (!alarmEndFlag && a == 0) {
                                alarmEndFlag = true;
                                soundOnTime = System.nanoTime();
                                alarmEnd();
                            } else if (!alarmStartFlag && (a > 0 && a < 6 * alarmStart)) {
                                alarmStartFlag = true;
                                if(a >= 6 * alarmStart - 0.5 && a <= 6 * alarmStart) {
                                    alarmBegin();
                                }
                            }
                            
                            if (led.getFillPaint() == ledColors[1]) {
                                if (!alarmStartFlag) {
                                    led.setFillPaint(ledColors[0]);
                                } else {
                                    led.setFillPaint(ledColors[2]);
                                }
                            } else {
                                led.setFillPaint(ledColors[1]);
                            }
                            
                            // Automatic turn-off after 1 minute.
                            if(alarmEndFlag && (TimeUnit.NANOSECONDS.toMinutes(now - soundOnTime) == 1)) {
                                turnOff();
                            }
                        }
                    }
                });
        time.pause();
    }

    private void turnOff() {
        defaultSound.stop();
        led.setFillPaint(ledColors[0]);
        reset = true;

        rotate(0);

        if (time.isRunning()) {
            time.stop();
        }
    }

    public void setAlarmStartMinute(int alarmStart) {
        this.alarmStart = alarmStart;
    }

    // In this demo, listeners are not implemented.
    private void alarmBegin() {
    // notify alarm listeners with an AlarmBegin event. 
    // defaultSound.play();
    }

    private void alarmEnd() {
        // notify alarm listeners with an AlarmEnd event. 
        defaultSound.loop();
    }

    private void makeScene(SGGroup root) {
        SGGroup g = new SGGroup();

        FXShape ocircle = new FXShape();
        ocircle.setShape(new Ellipse2D.Double(0, 0, 2 * r, 2 * r));
        ocircle.setAntialiasingHint(RenderingHints.VALUE_ANTIALIAS_ON);
        ocircle.setFillPaint(Color.white);
        ocircle.addMouseListener(new SGMouseAdapter() {

            @Override
            public void mouseDragged(java.awt.event.MouseEvent e, SGNode node) {
                int dx = e.getX() - cx;
                int dy = e.getY() - cy;
                // If we're running as an applet the frame will be null
                if (frame != null) {
                    frame.setLocation(frame.getX() + dx, frame.getY() + dy);
                }
            }

            @Override
            public void mousePressed(java.awt.event.MouseEvent e, SGNode node) {
                cx = e.getX();
                cy = e.getY();
            }
        });

        FXShape mcircle = new FXShape();
        mcircle.setShape(new Ellipse2D.Double(m, m, 2 * (r - m), 2 * (r - m)));
        mcircle.setAntialiasingHint(RenderingHints.VALUE_ANTIALIAS_ON);
        mcircle.setFillPaint(Color.lightGray);

        FXShape icircle = new FXShape();
        icircle.setShape(new Ellipse2D.Double(2 * m, 2 * m,
                2 * (r - 2 * m), 2 * (r - 2 * m)));
        icircle.setAntialiasingHint(RenderingHints.VALUE_ANTIALIAS_ON);
        icircle.setFillPaint(diskFillColor);

        g.add(ocircle);
        g.add(mcircle);

        // major ticks
        g.add(createTicks(12, (int) m / 4));

        // minor ticks.
        if (r > 59) {
            g.add(createTicks(60, (m <= 20) ? 1 : (int) m / 20));
        }
        g.add(icircle);
        g.add(createKnob());
        g.add(createReferenceDot());

        if (r > 59) {
            for (SGNode one : createLabels()) {
                g.add(one);
            }
        }

        root.add(g);
    }

    private void rotate(double theta) {
        AffineTransform at = new AffineTransform();
        at.rotate(theta, r, r);
        knobAffine.setAffine(at);
        refAffine.setAffine(at);
    }

    private void rotateBy(double theta) {
        AffineTransform at = new AffineTransform();
        at.rotate(theta, r, r);
        knobAffine.transformBy(at);
        refAffine.transformBy(at);
    }

    private double getKnobAngle() {
        Rectangle2D b = refAffine.getBounds();
        double cx = b.getX() + b.getWidth() / 2;
        double cy = b.getY() + b.getHeight() / 2;

        double a = Math.toDegrees(Math.atan2((r - cy), (r - cx)));
        if (a < 0) {
            a = a + 360;
        }
        a = a - 90;
        if (a < 0) {
            a = a + 360;
        }
        return a;
    }

    private SGNode createKnob() {
        SGGroup kg = new SGGroup();

        Ellipse2D.Double kn = new Ellipse2D.Double(r - m, 2 * m, 2 * m, 2 * (r - 2 * m));
        FXShape ks = new FXShape();
        ks.setShape(kn);
        ks.setFillPaint(knobFillColor);
        ks.setDrawPaint(Color.black);
        ks.setMode(SGShape.Mode.STROKE_FILL);
        ks.setDrawStroke(new BasicStroke(1));
        ks.setAntialiasingHint(RenderingHints.VALUE_ANTIALIAS_ON);
        kg.add(ks);

        Ellipse2D.Double ld = new Ellipse2D.Double(r - 0.4 * m, 3 * m, 0.8 * m, 0.8 * m);
        led = new FXShape();
        led.setShape(ld);
        led.setFillPaint(ledColors[0]);
        led.setDrawPaint(Color.black);
        led.setMode(SGShape.Mode.STROKE_FILL);
        led.setDrawStroke(new BasicStroke(1));
        led.setAntialiasingHint(RenderingHints.VALUE_ANTIALIAS_ON);
        led.addMouseListener(new SGMouseAdapter() {

            @Override
            public void mouseClicked(java.awt.event.MouseEvent e, SGNode node) {
                turnOff();
            }
        });
        kg.add(led);
        kg.setMouseBlocker(true);
        kg.addMouseListener(new SGMouseAdapter() {

            int sx = -1;
            int sy = -1;

            @Override
            public void mouseDragged(java.awt.event.MouseEvent e, SGNode node) {
                int cx = e.getX();
                int cy = e.getY();
                double theta = Math.atan2(r - cy, r - cx) - Math.atan2(r - sy, r - sx);
                rotateBy(theta);

                sx = cx;
                sy = cy;
            }

            @Override
            public void mousePressed(java.awt.event.MouseEvent e, SGNode node) {
                sx = e.getX();
                sy = e.getY();

                if (time.isRunning()) {
                    led.setFillPaint(ledColors[0]);
                    time.stop();
                    reset = true;
                }

                defaultSound.stop();
            }

            @Override
            public void mouseReleased(java.awt.event.MouseEvent e, SGNode node) {
                led.setFillPaint(ledColors[1]);
                sx = -1;
                sy = -1;
                time.start();
            }
        });

        knobAffine = SGTransform.createAffine(new AffineTransform(), kg);
        return knobAffine;
    }

    /**
     * A reference dot is a small red circle on the ticks. It is used to 
     * calculate the angle of the knob, which is then used to calculate the 
     * remaining time.
     * 
     * @return a red circle of the SceneGraph node
     */
    private SGNode createReferenceDot() {
        Ellipse2D.Double ref = new Ellipse2D.Double(r - 2, 1.5 * m, 4, 4);
        FXShape ks = new FXShape();
        ks.setShape(ref);
        ks.setDrawPaint(Color.red);
        ks.setMode(SGShape.Mode.STROKE);
        ks.setDrawStroke(new BasicStroke(1));
        ks.setAntialiasingHint(RenderingHints.VALUE_ANTIALIAS_ON);

        refAffine = SGTransform.createAffine(new AffineTransform(), ks);
        return refAffine;
    }

    private SGNode createTicks(int num, int strokeWidth) {
        GeneralPath p = new GeneralPath();
        double[] x = new double[num];
        double[] y = new double[num];
        for (int i = 0; i < num; i++) {
            double radians = Math.toRadians(360 / num * i - 90);
            x[i] = (r - 1.3 * m) * Math.cos(radians);
            y[i] = (r - 1.3 * m) * Math.sin(radians);
        }

        for (int i = 0; i < num / 2; i++) {
            p.moveTo(x[i], y[i]);
            p.lineTo(x[i + num / 2], y[i + num / 2]);
        }

        FXShape tick = new FXShape();
        tick.setShape(p);
        tick.setFillPaint(Color.black);
        tick.setDrawPaint(Color.black);
        tick.setMode(SGShape.Mode.STROKE_FILL);
        tick.setDrawStroke(new BasicStroke(strokeWidth));
        tick.setAntialiasingHint(RenderingHints.VALUE_ANTIALIAS_ON);
        SGTransform.Translate trans = SGTransform.createTranslation(r, r, tick);
        return trans;
    }

    private SGNode[] createLabels() {
        SGTransform.Translate[] labels = new SGTransform.Translate[12];
        for (int i = 0; i < 12; i++) {
            FXText fxText = new FXText();
            if (i == 0) {
                fxText.setFillPaint(Color.red);
            } else {
                fxText.setFillPaint(Color.black);
            }
            fxText.setAntialiasingHint(RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
            fxText.setFont(new Font("SansSerif", Font.BOLD, fontSize));

            //((FXNode)fxText).setVerticalAlignment(3); TOP
            //fxText.setHorizontalTextAlignment(HAlign.CENTER);
            ((SGText)(fxText.getLeaf())).setVerticalAlignment(VAlign.BOTTOM);
            ((SGText)(fxText.getLeaf())).setHorizontalAlignment(HAlign.LEFT);
            fxText.setText(5 * i + "");

            double radians = Math.toRadians(30 * i - 90);
            double x = (r - m / 2) * Math.cos(radians);
            double y = (r - m / 2) * Math.sin(radians);
            double width = 20;  // width of fxText
            double height = 15; // height of fxText
            double dx = (width/2)*Math.cos(radians) - (width/2);
            double dy = (-height/2)*Math.sin(radians) - (height/2);
            fxText.setLocation(new Point2D.Double(x + dx, y - dy));
            SGTransform.Translate trans = SGTransform.createTranslation(r, r, fxText);
            labels[i] = trans;
        }
        return labels;
    }

    private void invalidateScene() {
        SGGroup root = new SGGroup();
        makeScene(root);
        SGTransform uberRoot = SGTransform.createTranslation(10, 8, root);
        setScene(uberRoot);
    }

    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Runnable() {

            @Override
            public void run() {
                try {
                    String name = UIManager.getSystemLookAndFeelClassName();
                    UIManager.setLookAndFeel(name);
                } catch (Exception ignore) {
                }

                EggTimer app = new EggTimer();
                frame = new JFrame("SceneGraph Demo");
                frame.setUndecorated(true);
                frame.setBackground(new Color(1.0f, 1.0f, 1.0f, 0.5f));
                frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
                frame.add(app, BorderLayout.CENTER);
                frame.pack();
                //frame.setSize(300,400);
                frame.setLocationRelativeTo(null);
                frame.setVisible(true);

                final JPopupMenu popup = new JPopupMenu();
                JMenuItem hide = new JMenuItem("Hide");
                hide.setMnemonic(KeyEvent.VK_H);
                hide.addActionListener(new ActionListener() {

                    @Override
                    public void actionPerformed(ActionEvent e) {
                        frame.setState(Frame.ICONIFIED);
                    }
                });
                popup.add(hide);

                JCheckBoxMenuItem ontop = new JCheckBoxMenuItem("Always On Top");
                ontop.setMnemonic(KeyEvent.VK_A);
                ontop.addActionListener(new ActionListener() {

                    @Override
                    public void actionPerformed(ActionEvent e) {
                        JCheckBoxMenuItem menu = (JCheckBoxMenuItem) e.getSource();
                        frame.setAlwaysOnTop(menu.getState());
                    }
                });
                popup.add(ontop);
                popup.addSeparator();

                JMenuItem close = new JMenuItem("Close");
                close.setMnemonic(KeyEvent.VK_C);
                close.addActionListener(new ActionListener() {

                    @Override
                    public void actionPerformed(ActionEvent e) {
                        frame.dispose();
                        System.exit(0);
                    }
                });
                popup.add(close);

                app.addMouseListener(new MouseAdapter() {

                    @Override
                    public void mousePressed(MouseEvent e) {
                        if (e.isPopupTrigger()) {
                            popup.show(e.getComponent(), e.getX(), e.getY());
                        }
                    }

                    @Override
                    public void mouseReleased(MouseEvent e) {
                        if (e.isPopupTrigger()) {
                            popup.show(e.getComponent(), e.getX(), e.getY());
                        }
                    }
                });
            }
        });
    }

    @Override
    public Dimension getPreferredSize() {
        return new Dimension((int) r * 2 +25, (int) r * 2 + 25);
    }
}
